import re

class Hla():
    def Reset(self):
        self.startTime = -1
        self.dataStart = -1
        self.address = 0
        self.endTime = -1
        self.data = []

    ''' Utility routines
    '''

    def MakeTextFrame(self, start, stop, text):
        return {
            'type': 'text',
            'start_time': start, 'end_time': stop,
            'data': {'text': text}
            }

    def MakeListFrame(self, start, stop, list):
        listStr = ", ".join((str(item) for item in list))
        return {
            'type': 'text',
            'start_time': start, 'end_time': stop,
            'data': {'text': listStr}
            }

    ''' Register content decoders
    '''

    def DecodeAlerts(self):
        kAlertBits = 0x8FFF
        newAlerts = kAlertBits & (self.data[0][0] | (self.data[1][0] << 8))
        isWrite = not (self.address & 1)

        if not newAlerts:
            setAlerts = ["none"]
        elif newAlerts == kAlertBits:
            setAlerts = ["all"]
        else:
            setAlerts = \
                [entry[1] for entry in self.alertBits if newAlerts & entry[0]]

        if isWrite:
            setAlerts[0] = "clear " + setAlerts[0]

        return setAlerts

    def DecodeAlertMask(self):
        kAlertBits = 0x8FFF
        kInterestingBits = 0x8E7F
        isWrite = not (self.address & 1)
        newAlerts = self.data[0][0] | (self.data[1][0] << 8)
        changedAlerts = self.lastAlerts ^ newAlerts

        if not changedAlerts and not newAlerts:
            changedAlerts = kAlertBits

        setAlertBits = changedAlerts & newAlerts
        resetAlertBits = changedAlerts & ~newAlerts
        self.lastAlerts = newAlerts

        if ((newAlerts & kInterestingBits) == kInterestingBits) \
           or setAlertBits == kAlertBits:
            return ["enable all"]

        if not newAlerts or resetAlertBits == kAlertBits:
            return ["disable all"]

        setAlerts = \
            [entry[1] for entry in self.alertBits if setAlertBits & entry[0]]
        resetAlerts = \
            [entry[1] for entry in self.alertBits if resetAlertBits & entry[0]]

        if setAlertBits:
            setAlerts[0] = "enable " + setAlerts[0]

        if resetAlertBits:
            resetAlerts[0] = "disable " + resetAlerts[0]

        return setAlerts + resetAlerts

    def DecodePowerAlerts(self):
        byte = self.data[0][0]
        flags = []

        if byte & 0x80: flags.append("Dbg Acc")
        if byte & 0x40: flags.append("TCPC Int")
        if byte & 0x20: flags.append("Src HiVBus")
        if byte & 0x10: flags.append("Src VBus")
        if byte & 0x08: flags.append("Det VBus")
        if byte & 0x04: flags.append("VBus Pres")
        if byte & 0x02: flags.append("VConn Pres")
        if byte & 0x01: flags.append("Sinking VBus")

        if not (self.address & 1) and self.registerNum == 0x14:
            flags[0] = "Enable " + flags[0]

        return flags

    def DecodeFault(self):
        byte = self.data[0][0]
        flags = []

        if (byte & 0x7f) == 0x7f:
            flags = ["all"]

        elif (byte & 0x7f) == 0:
            flags = ["none"]

        else:
            if byte & 0x40: flags.append("VBus Off")
            if byte & 0x20: flags.append("Auto Dis")
            if byte & 0x10: flags.append("Dis fail")
            if byte & 0x08: flags.append("VBus OCP")
            if byte & 0x04: flags.append("VBus OVP")
            if byte & 0x02: flags.append("VConn OCP")
            if byte & 0x01: flags.append("Tx underflow")

        if not (self.address & 1):
            if self.registerNum == 0x1F: flags[0] = "clear " + flags[0]
            else: flags[0] = "Enable " + flags[0]

        return flags

    def DecodeFaultCtrl(self):
        byte = self.data[0][0]
        flags = []

        if byte & 0x08: flags.append("VBus Dis timer on")
        else: flags.append("VBus Dis timer off")
        return flags

    def DecodeTCPC(self):
        byte = self.data[0][0]

        if byte & 1: flags = ["Flip"]
        else: flags = ["Normal"]

        return flags;

    def DecodeRole(self):
        ccState = {
            0: "Ra",
            1: "Rp",
            2: "Rd",
            3: "Open"
            }

        byte = self.data[0][0]
        flags = [];

        self.isSource = (byte & 0x03) == 0x01 and (byte & 0x0C) == 0x040
        flags.append(ccState[byte & 0x03] + "1")
        flags.append(ccState[(byte >> 2) & 0x03] + "2")
        return flags

    def DecodePowerCtrl(self):
        byte = self.data[0][0]

        flags = [];
        if byte & 0x01: flags.append("VConn On")
        else: flags.append("VConn Off")

        if byte & 0x04: flags.append("Dis")
        if byte & 0x08: flags.append("Blead")
        if byte & 0x10: flags.append("Auto")
        if byte & 0x20: flags.append("No VAlarm")
        if byte & 0x40: flags.append("VBus Mon")

        return flags;

    def DecodeCCState(self):
        ccSrcState = {
            0: "Open",
            1: "Ra",
            2: "Rd",
            3: "?",
            }
        ccSinkState = {
            0: "Open",
            1: "Def",
            2: "1.5A",
            3: "3.0A",
            }

        byte = self.data[0][0]
        flags = [];

        if self.isSource:
            flags.append("CC1 " + ccSrcState[byte & 0x03])
            flags.append("CC2 " + ccSrcState[(byte >> 2) & 0x03])
        else:
            flags.append("CC1 " + ccSinkState[byte & 0x03])
            flags.append("CC2 " + ccSinkState[(byte >> 2) & 0x03])

        return flags

    def DecodeMsgHeaderInfo(self):
        byte = self.data[0][0]
        flags = []

        if byte & 0x01: flags.append("Src")
        else: flags.append("Sink")

        flags.append(["R1", "R2", "R3", "x"][(byte & 0x06) >> 1])

        if byte & 0x08: flags.append("DFP")
        else: flags.append("UFP")

        if byte & 0x10: flags.append("Cable")

        return flags

    def DecodeRxDetect(self):
        byte = self.data[0][0]

        flags = [];
        if byte & 0x01: flags.append("SOP")
        if byte & 0x02: flags.append("SOPP")
        if byte & 0x04: flags.append("SOPPP")
        if byte & 0x08: flags.append("SOP DBGP")
        if byte & 0x10: flags.append("SOP DBGPP")
        if byte & 0x20: flags.append("Hard Reset")
        if byte & 0x40: flags.append("Cable Reset")

        return flags;

    def DecodeRxType(self):
        byte = self.data[0][0]
        flags = [];

        flags.append([
            "SOP",
            "SOP'",
            "SOP''",
            "SOP_DBG'",
            "SOP_DBG''",
            "Cable Reset",
            "BIST",
            ][byte & 0x07])

        return flags

    def DecodeTx(self):
        byte = self.data[0][0]
        flags = [];

        flags.append("Retry " + str((byte >> 4) & 0x03))
        flags.append([
            "SOP",
            "SOP'",
            "SOP''",
            "SOP_DBG'",
            "SOP_DBG''",
            "Hard Reset",
            "Cable Reset",
            "BIST",
            ][byte & 0x07])

        return flags

    def DecodeCount(self):
        count = self.data[0][0]
        return [str(count)]

    def DecodeFixed(self, word):
        flags = []

        if (word >> 29) & 1: flags.append("DR Power")
        if (word >> 28) & 1: flags.append("Suspend")
        if (word >> 27) & 1: flags.append("Unconstrained")
        if (word >> 26) & 1: flags.append("USB")
        if (word >> 25) & 1: flags.append("DR Data")

        overCurrent = (word >> 20) & 0x03
        if overCurrent:
            flags.append(["Peak low", "Peak mid", "Peak high"][overCurrent - 1])
        flags.append(str(((word >> 10) & 0x3ff) * 0.05) + "V")
        flags.append(str((word & 0x3ff) * 0.01) + "A")
        return flags

    def DecodeBatt(self, word):
        flags = []
        flags.append("Min " + str(((word >> 20) & 0x3ff) * 0.05) + "V")
        flags.append("Max " + str(((word >> 10) & 0x3ff) * 0.05) + "V")
        flags.append(str((word & 0x3ff) * 0.25) + "W")
        return flags

    def DecodeVariable(self, word):
        flags = []
        flags.append("Min " + str(((word >> 20) & 0x3ff) * 0.05) + "V")
        flags.append("Max " + str(((word >> 10) & 0x3ff) * 0.05) + "V")
        flags.append(str((word & 0x3ff) * 0.01) + "A")
        return flags

    def DecodeSrcCaps(self, word):
        capsType = word >> 30

        if capsType == 3:
            return ["Bad caps"]

        return self.decodeCapsDispatch[capsType](word)

    def DecodeRequest(self, word):
        flags = ["Cap Num " + str((word >> 28) & 0x07)]

        if (word >> 27) & 1: flags.append("Give back")
        if (word >> 26) & 1: flags.append("Mismatch")
        if (word >> 25) & 1: flags.append("USB")
        if (word >> 24) & 1: flags.append("No Suspend")

        flags.append(str(((word >> 10) & 0x3FF) * 0.01) + "A")
        flags.append("max " + str((word & 0x3FF) * 0.01) + "A")

        return flags

    def DecodeVDM(self, word):
        flags = []

        if re.match("RxPDO [14]", self.register):
            if self.register == "RxPDO 1":
                # ID Header
                self.cableType = [
                    "undef", "Res1", "Res2", "Passive", "Active"
                    ][(word >> 27) & 0x07]
                return [self.cableType]

            if self.register == "RxPDO 4":
                # Cable VDO
                flags.append(
                    ["USB A", "USB B", "USB C", "Captive"][(word >> 18) & 0x03]
                    )

                flags.append("<= " + str((word >> 13) & 0x0f) + "m")
                if (word >> 11) & 0x03: flags.append("Need VConn")
                flags.append(["?A", "3A", "5A"][(word >> 5) & 0x03])
                if (word >> 4) & 0x01: flags.append("Has VBus")

        elif re.match("TxPDO 0", self.register):
            cmd = word & 0x0000000f

            if cmd > 15: flags.append("Bad cmd " + str(cmd))
            elif cmd > 6: flags.append("SVID cmd " + str(cmd))
            else: flags.append(
                [
                "BAD!",
                "Discover Id",
                "Discover SVIDs",
                "Discover Modes",
                "Enter",
                "Exit",
                "Attention"
                ][cmd]
                )

            flags.append(["REQ", "ACK", "NAK", "BUSY"][(word >> 6) & 0x03])

        if flags:
            return flags;

        return [format(word, "#010x")]

    def DecodeWord(self):
        byte0 = self.data[0][0]
        byte1 = self.data[1][0]
        byte2 = self.data[2][0]
        byte3 = self.data[3][0]
        word = byte0 | (byte1 << 8) | (byte2 << 16) | (byte3 << 24)

        if self.messageType in self.decodeWordDispatch:
            return self.decodeWordDispatch[self.messageType](word)

        return [format(word, "#010x")]

    def DecodeWords(self):
        flags = []
        nPDO = 0

        while len(self.data) >= 4:
            block = self.DecodeWord()
            if nPDO: block[0] = " PDO" + str(nPDO) + ": " + block[0]
            nPDO += 1
            flags += block
            self.data = self.data[4:]

        return flags

    def DecodeMsgHeader(self):
        h0 = self.data[0][0]
        h1 = self.data[1][0]
        flags = []

        nPDO = (h1 >> 4) & 0x07
        if nPDO: flags.append(str(nPDO) + " PDOs")

        msgId = str((h1 >> 1) & 0x03)
        flags.append("ID " + msgId)

        msgType = h0 & 0x0F

        if nPDO:
            # Data message
            self.messageType = {
                0: "!bad!",
                1: "SrcCaps",
                2: "Request",
                3: "BIST",
                4: "SinkCaps",
                15: "VDM"
                }[msgType]
        else:
            # Control message
            self.messageType = {
                0: "!bad!",
                1: "GoodCRC",
                2: "GotoMin",
                3: "Accept",
                4: "Reject",
                5: "Ping",
                6: "PS_Rdy",
                7: "GetSrcCaps",
                8: "GetSinkCap",
                9: "DRSwap",
                10: "PRSwap",
                11: "VConnSwp",
                12: "Wait",
                13: "Reset"
                }[msgType]

        flags.append(self.messageType)
        return flags

    def DecodeVBusV(self):
        word = self.data[0][0] | (self.data[1][0] << 8)
        flags = [str(word * 0.025) + "V"]

        return flags

    def DecodeGenCtl(self):
        byte = self.data[0][0]
        flags = []

        if byte & 0x20: flags.append("SW Reset")
        else: flags.append("Ready")

        return flags

    ''' I2C part handlers
    '''

    def DecodeGPIOA(self):
        byte = self.data[0][0]
        flags = []

        if self.address & 1:
            if byte & 0x01: flags.append("PGood")
            else: flags.append("PBad")

            if byte & 0x40: flags.append("VBus good")
            elif byte & 0x20: flags.append("VBus bad")

            if byte & 0x08: flags.append("Alert I")
            if byte & 0x10: flags.append("USBFault")

        if byte & 0x02: flags.append("Discharge")

        if byte & 0x04: flags.append("Pwr on")
        else: flags.append("Pwr off")

        if byte & 0x20: flags.append("VBus off")
        else: flags.append("VBus on")

        if byte & 0x80: flags.append("Mux flip")

        return flags

    def DecodeGPIOB(self):
        byte = self.data[0][0]
        flags = []

        if byte & 0x01: flags.append("Mux off")
        else: flags.append("Mux on")

        if byte & 0x02: flags.append("Slew limit")
        else: flags.append("No slew limit")

        if byte & 0x04: flags.append("9V")
        if byte & 0x08: flags.append("15V")
        if byte & 0x10: flags.append("20V")
        if not (byte & 0x1C): flags.append("5V")
        if byte & 0x20: flags.append("GB5")
        if byte & 0x40: flags.append("GB6")
        if byte & 0x80: flags.append("GB7")

        return flags

    def DecodeIDAC(self):
        kTotalSenseR = 0.005;
        kSenseGain = 50.0;
        kDACRangeV = 2.048;
        kDAC_LSB = kDACRangeV / 4096;
        kScaleRefA2DAC = kTotalSenseR * kSenseGain / kDAC_LSB;

        rawDAC = ((self.data[1][0] & 0x7F) << 8) + self.data[2][0]
        return "Port " + str((self.data[0][0] & 0x06) >> 1) \
            + " = " + format(rawDAC / kScaleRefA2DAC, ".1f") + "A"

    def DecodeIO8(self):
        byte = self.data[0][0]
        text = ""

        if self.address & 1: # It's a read
            if byte & 0x01: text += "USB Fault, "
            if byte & 0x08: text += "VBus SW PGood, "
        else: # It's a write
            if ~byte & 0x02: text += "EMux En, "
            if byte & 0x04: text += "EMux Sel, "
            if ~byte & 0x10: text += "!VBus SW En, "
            if byte & 0x20: text += "Boost On"
            else: text += "Boost Off"

        return text

    def HandleStart(self, params):
        self.Reset()

    def HandleAddress(self, params):
        self.address = params["data"]["address"][0]
        self.device = ""

        if self.address == 0x40 or self.address == 0x41:
            self.device = "PD"

        if self.address == 0xC0 or self.address == 0xC1:
            self.device = "IDAC"

        if self.address >= 0x42 and self.address <= 0x49:
            self.device = "I/O16"

        if self.address == 0x78 or self.address == 0x79:
            self.device = "I/O8"

        if self.device:
            text = self.device
        else:
            text =  "a" + str(self.address)

        if not (self.address & 1):
            # Reset register on a write transaction
            self.register = ""
            self.registerNum = -1

        return self.MakeTextFrame(params['start_time'], params['end_time'], text)

    def HandleData(self, params):
        self.dataEnd = params['end_time']

        if self.device in self.dispatchLU \
            and not (self.address & 1) and not self.register:
            # It's a write transaction for a decoded register based device
            # The first data byte is a register address
            self.registerNum = params["data"]["data"][0]

            if self.device in self.regSelLU:
                regLU = self.regSelLU[self.device]

                if self.registerNum in regLU:
                    self.register = regLU[self.registerNum]

            if not self.register:
                self.register = format(self.registerNum, "#04x")

            return self.MakeTextFrame(params['start_time'], self.dataEnd, self.register)

        if not self.data:
            self.dataStart = params['start_time']

        self.data.append(params["data"]["data"])

    def HandleStop(self, params):
        # Dump the transaction's data
        if not self.data or self.dataStart < 0:
            return # No data, nothing to dump

        self.startTime = -1 # Done with this transaction
        decodeTable = {}

        if self.device in self.dispatchLU:
            decodeTable = self.dispatchLU[self.device]
        else:
            if self.device == "IDAC":
                text = self.DecodeIDAC()
            elif self.device == "I/O8":
                text = self.DecodeIO8()
            else:
                text = "!Can't decode " + self.device

            return self.MakeTextFrame(self.dataStart, self.dataEnd, text)

        if not decodeTable or not (self.registerNum in decodeTable):
            text = " ".join(format(value[0], "#04x") for value in self.data)
            text = text
            return self.MakeTextFrame(self.dataStart, self.dataEnd, text)

        result = decodeTable[self.registerNum]()

        if self.address & 1 and result:
            result[0] = self.register + ": " + result[0]
            match = re.match("RxPDO (\d+)", self.register)

            if match:
                self.register = "RxPDO " + str(int(match.group(1)) + 1)

        return self.MakeListFrame(self.dataStart, self.dataEnd, result)

    ''' Logic2 API entry points and class construction
    '''

    def __init__(self):
        self.isSource = False
        self.Reset()
        self.PDRegisters = {
            0x10: "AlertState",
            0x12: "Alert&",
            0x14: "Power&",
            0x15: "Fault&",
            0x19: "TCPC",
            0x1A: "Role",
            0x1C: "Power",
            0x1D: "CCState",
            0x1F: "FaultState",
            0x2E: "MsgInfo",
            0x2F: "RxDet",
            0x30: "RxN",
            0x31: "RxType",
            0x32: "RxHdr",
            0x34: "RxPDO 0",
            0x38: "RxPDO 1",
            0x3C: "RxPDO 2",
            0x40: "RxPDO 3",
            0x44: "RxPDO 4",
            0x48: "RxPDO 5",
            0x4C: "RxPDO 6",
            0x50: "TxSend",
            0x51: "TxN",
            0x52: "TxHdr",
            0x54: "TxPDO 0",
            0x58: "TxPDO 1",
            0x5C: "TxPDO 2",
            0x60: "TxPDO 3",
            0x64: "TxPDO 4",
            0x68: "TxPDO 5",
            0x6C: "TxPDO 6",
            0x70: "kVBusV0",
            0x76: "VAlarm!",
            0x90: "VendInt",
            0x92: "VendInt&",
            0x94: "GenCtl",
            0x98: "VConn"
            }
        self.IORegisters = {
            0x00: "A Dir",
            0x01: "B Dir",
            0x04: "A Int enab",
            0x05: "B Int enab",
            0x12: "GPIOA",
            0x13: "GPIOB",
            }
        self.alertBits = [
            [0x0001, "CC"],
            [0x0002, "Pwr"],
            [0x0004, "RxSOP"],
            [0x0008, "RxRst"],
            [0x0010, "TxFail"],
            [0x0020, "TxDrop"],
            [0x0040, "TxSOP"],
            [0x0080, "VHi"],
            [0x0100, "VLo"],
            [0x0200, "Fault"],
            [0x0400, "RxOvf"],
            [0x0800, "VDis"],
            [0x8000, "VIrq"]
        ]
        self.decodePDDispatch = {
            0x10: self.DecodeAlerts,
            0x12: self.DecodeAlertMask,
            0x14: self.DecodePowerAlerts,
            0x15: self.DecodeFault,
            0x19: self.DecodeTCPC,
            0x1A: self.DecodeRole,
            0x1B: self.DecodeFaultCtrl,
            0x1c: self.DecodePowerCtrl,
            0x1D: self.DecodeCCState,
            0x1f: self.DecodeFault,
            0x2E: self.DecodeMsgHeaderInfo,
            0x2F: self.DecodeRxDetect,
            0x30: self.DecodeCount,
            0x31: self.DecodeRxType,
            0x32: self.DecodeMsgHeader,
            0x34: self.DecodeWords,
            0x38: self.DecodeWords,
            0x3C: self.DecodeWords,
            0x40: self.DecodeWords,
            0x44: self.DecodeWords,
            0x48: self.DecodeWords,
            0x4C: self.DecodeWords,
            0x50: self.DecodeTx,
            0x51: self.DecodeCount,
            0x52: self.DecodeMsgHeader,
            0x54: self.DecodeWord,
            0x58: self.DecodeWord,
            0x5C: self.DecodeWord,
            0x60: self.DecodeWord,
            0x64: self.DecodeWord,
            0x68: self.DecodeWord,
            0x6C: self.DecodeWord,
            0x70: self.DecodeVBusV,
            0x94: self.DecodeGenCtl,
            }
        self.decodeIO16Dispatch = {
            0x12: self.DecodeGPIOA,
            0x13: self.DecodeGPIOB,
            }
        self.decodeWordDispatch = {
            "SrcCaps": self.DecodeSrcCaps,
            "Request": self.DecodeRequest,
            "VDM": self.DecodeVDM
            }
        self.decodeCapsDispatch = [
            self.DecodeFixed,
            self.DecodeBatt,
            self.DecodeVariable
            ]
        self.handlerDispatch = {
            "start": self.HandleStart,
            "address": self.HandleAddress,
            "data": self.HandleData,
            "stop": self.HandleStop,
            }
        self.dispatchLU = {
            "PD": self.decodePDDispatch,
            "I/O16": self.decodeIO16Dispatch,
            }
        self.regSelLU = {
            "PD": self.PDRegisters,
            "I/O16": self.IORegisters,
            }
        self.lastAlerts = 0
        self.device = ""

    def get_capabilities(self):
        pass

    def set_settings(self, settings):
        return {
            "result_types": {
                "text": {"format": "{{{data.text}}}"},
            }   }

    def decode(self, params):
        if params['type'] in self.handlerDispatch:
            # Handle the I2C phase data
            return self.handlerDispatch[params['type']](params)

        # We don't handle this I2C phase type!
        new_frame = {
            'type': 'text',
            'start_time': params['end_time'],
            'end_time': params['end_time'] + 0.00002,
            'data': {'text': '!' + params['type']}
            }
        return new_frame
